/**
 * Data_info.cc
 *
 * Version :   0.2
 * Author  :   Junming Yin
 * 
 * (c) UC Berkeley, 2010
 * 
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <cctype>	        // for isdigit
#include <time.h>               // for time(NULL)
#include <getopt.h>	        // for getopt_long and getopt_long_only
#include <gsl/gsl_randist.h>    // for gsl_ran_shuffle
#include <string.h>             // for ctype strings

#include "Data_info.h"


// test whether the input is binary
#define isBinary(ch) ( ((ch) == 0 || (ch) == 1) )
// test whether the argument for -mle is valid
#define isMLEValid(arg) ( strlen((arg)) == 1 && isdigit(*(arg)) && isBinary(atoi((arg))) )


using namespace std;


/****************************
     gsl global generator
****************************/

const gsl_rng_type* T;
gsl_rng* rg;                             

/****************************
 ****************************/


// default constructor with explicit initializers
Data_info::Data_info(): N(0), L(0), divisor(1), nOrder(0), nRep(0), theta(0) { }


// constructor with arguments
Data_info::Data_info(char *in_file, char* order_file, int n_order, int n_rep, double mut_rate) {  

  // read the in_file that contains the data
  ifstream datafile(in_file);
  if(!datafile.is_open())
    {
      cout << "Could not open file "<< in_file << endl;
      exit(1);
    }


  /****************************
   initialize N, L and divisor
  ****************************/

  // read the first line
  char line[MAX_LENGTH];
  datafile.getline(line, MAX_LENGTH);
  if( sscanf(line, "%d %d %ld", &N, &L, &divisor) < 2 )
    {
      cout << in_file << " has wrong format! The first row should contain number of haplotypes and number of SNPs." << endl;
      exit(1);
    }
  if(divisor <= 0)
    divisor = 1;
  

  /********************************
    initialize nOrder, nRep, theta
  *********************************/

  nOrder = n_order;
  nRep = n_rep;

  // user specifies the mutation rate 
  if (mut_rate != 0.0)  
    theta = mut_rate;
  // Watterson's estimator (mutation rate per site)
  else { 
    double temp = 0.0;
      
    for(int i=1; i<N; i++)
      temp += 1.0 /(double) i;
    
    theta = 1.0/temp;
  }

  /****************************
       initialize loci
  ****************************/

  // get the second loci-position line
  datafile.getline(line, MAX_LENGTH);
  stringstream strStream(line);
  if(strStream.str().size() == 0) {
    cout << in_file << " has wrong format! The second row should contain the positions of SNPs." << endl;
    exit(1);
  }
  vector<double> v_dist;
  double val;
  while( strStream >> val) 
    v_dist.push_back(val);
  if ((int) v_dist.size() != L) {
    cout << in_file << " has wrong format! The second row should have " << L << "entries" << endl;
    exit(1);
  }
  strStream.clear();

  // compute the inter-loci distance
  loci.resize(L);
  loci[0].index = 0;
  loci[0].dist = 0.0;

  for(int i = 1; i < L;i++) {
    loci[i].index = i;
    loci[i].dist = (v_dist[i] - v_dist[i-1])/(double) divisor;
  }

  /****************************
    initialize haplotypes
  ****************************/

  // read the haplotype matrix
  int nRow, nCol;
  readMatrix(in_file, datafile, haplotypes, nRow, nCol);
  if (nRow != N || nCol != L) {
    cout << in_file << " has wrong format! The number of haplotypes should be " << N << " and the number of SNPS should be " << L << endl;
    exit(1);
  }
  

  /****************************
      initialize orderings
  ****************************/

  // user specifies the order_file, then ignore -nOrder, -nRep and -seed options
  if (order_file != '\0') {
    ifstream orderfile(order_file);
    if(!orderfile.is_open())
      {
        cout << "Could not open ordering file "<< order_file << endl;
        exit(1);
      }
    // there is only one repetition 
    nRep = 1;
    orderings.resize(nRep);

    // read the ordering matrix
    int n;
    readMatrix(order_file, orderfile, orderings[0], nOrder, n);
    if (n != N)
      {
        cout << order_file << " has wrong format! Each row should contain " << N << " entries." << endl;   
        exit(1);
      }
  }
  // random orderings, using -nOrder (n_order), -nRep(n_rep) and -seed(seed) options
  else{
    int a[N];
    for (int i = 0; i < N; i++)
      a[i] = i + 1;

    orderings.resize(nRep);
    for (int r = 0; r < nRep; r++) {
      intMatrix ord;
      ord.resize(nOrder);
      for(int n = 0; n < nOrder; n++) {
	// random shuffle a[] in place
	gsl_ran_shuffle (rg, a, N, sizeof (int));	
	ord[n].resize(N);
	for(int i = 0; i < N; i++) 
	  ord[n][i] = a[i];
      } // end n
      orderings[r] = ord;
    } // end r
  } // end else
  

  /****************************
   finish initializing members
  ****************************/
  datafile.close();
}

	
void Data_info::readMatrix(char *file, ifstream& fin, intMatrix& matrix, int& nRow, int& nCol)
{
  // read intMatrix into matrix from ifstream fin, which is initialized from "file"
  // return the intMatrix, number of rows and columns

  matrix.clear();
  char line[MAX_LENGTH];
  
  // read the first line
  fin.getline(line,MAX_LENGTH);
  stringstream strStream(line);
  if(strStream.str().size() == 0){  // empty first line
    cout << file << " is an empty file/matrix" << endl;
    exit(1);
  }
  // the first line is non-empty
  matrix.resize(1);
  int val; 
  nRow = 0;
  while( strStream >> val)
    matrix[nRow].push_back(val);
  // set the number of columns
  nCol = matrix[nRow].size();
  strStream.clear();
  nRow++;

  // read the remaining lines
  while( fin.getline(line,MAX_LENGTH) ) {
    stringstream strStream(line);
    if(strStream.str().size() > 0) {
      matrix.resize(matrix.size() + 1);
      while( strStream >> val) 
	matrix[nRow].push_back(val);

      if((int) matrix[nRow].size() != nCol) {
	cout << file << " has bad format! Row " << nRow << " should contain " << nCol << " entries." << endl;
	exit(1);
      }
    }
    strStream.clear();
    nRow++;
  }

  fin.close();
}


/*******************************
 (defualt) command line options
*******************************/
string command;
char *in_file = '\0';
char *out_file = '\0';

bool v_flag = 0;
unsigned long seed = 0;

char *order_file = '\0';
int n_order = 20;
int n_rep = 5;

bool mle_flag = 0;
bool rho_flag = 0;
bool gamma_flag = 0;
bool tl_flag = 0;

double initial_theta = 0.0;
double initial_rho = 0.0;
double initial_gamma = 0.0;
double initial_tl= 0.0;

// long options for parsing command line
static struct option long_options[] =
  {
    {"in",      required_argument, 0, 'i'},	// data file to use
    {"out",     required_argument, 0, 'o'},	// file to save the results

    {"verbose", no_argument,       0, 'v'},	// verbose output
    {"seed",    required_argument, 0, 's'},	// random seed, default value: time(NULL)

    {"order",   required_argument, 0, 'd'},	// order file to use
    {"nOrder",  required_argument, 0, 'n'},	// number of orderings to use, default value: 20
    {"rep",     required_argument, 0, 'p'},	// number of repetitions (using different random orderings), default value: 5
    {"mle",     required_argument, 0, 'm'},	// need three arguments

    {"theta",   required_argument, 0, 't'},	// initial mutation rate, default value: Watterson's estimator
    {"rho",     required_argument, 0, 'r'},	// initial recombination rate
    {"gamma",   required_argument, 0, 'g'},	// initial gene conversion rate
    {"tl",      required_argument, 0, 'l'},	// initial mean track length of gene conversion, 1/lambda

    {"help",    no_argument,       0, 'h'},	// this help

    {0, 0, 0, 0}				// terminate the array with an element containing all zeros
  };


int parseCommandline(int argc, char** argv, int flag)
{
  while (1)
    {
      /* getopt_long_only stores the option index here. */
      int option_index = 0;
     
      int c = getopt_long_only (argc, argv, "i:o:vs:d:n:r:m:t:p:g:l:h",
				long_options, &option_index);
         
      /* Detect the end of the options. */
      if (c == -1)
        break;
     
      switch (c)
        {
        case 'i':
          in_file = optarg;
          break;
     

        case 'o':
	  out_file = optarg;
          break;
     

        case 'v':
	  v_flag = 1;
          break;
	  

        case 's':
	  seed = atoi(optarg);
          break;

     
        case 'd':
	  order_file = optarg;
          break;
     
	  
        case 'n':
	  if (atoi(optarg) > 0)
	    n_order = atoi(optarg);
	  else {
	    fprintf (stderr, "Option -nOrder requires a positive argument.\n");	    
	    return 1;
	  }
          break;


        case 'p':
	  if (atoi(optarg) > 0)
	    n_rep = atoi(optarg);
	  else {
	    fprintf (stderr, "Option -rep requires a positive argument.\n");	    
	    return 1;
	  }
          break;

	
        case 'm':
          if(optind + 1 > argc - 1)	{		// less than three arguments
            fprintf (stderr, "Option -mle requires three binary arguments.\n");
            return 1;
          }
          else{					        // has three arguments
	    // has three arguments, and all of them are binary (valid)
	    if( isMLEValid(optarg) && isMLEValid(argv[optind]) && isMLEValid(argv[optind+1]) ) {	
	      rho_flag = atoi(optarg);
	      gamma_flag = atoi(argv[optind++]);
	      tl_flag = atoi(argv[optind++]);
	    }
            else {        // has three arguments, but at least one of them is not binary
	      fprintf (stderr, "Option -mle requires three binary arguments.\n");	    
	      return 1;
            }
          }
          break;


        case 't':
	  if (atof(optarg) > 0)
	    initial_theta = atof(optarg);
	  else {
	    fprintf (stderr, "Option -theta requires a positive argument.\n");	    
	    return 1;
	  }
          break;
     


        case 'r':
	  if (atof(optarg) > 0)
	    initial_rho = atof(optarg);
	  else {
	    fprintf (stderr, "Option -rho requires a positive argument.\n");	    
	    return 1;
	  }
          break;



        case 'g':
	  if (atof(optarg) > 0)
	    initial_gamma = atof(optarg);
	  else {
	    fprintf (stderr, "Option -gamma requires a positive argument.\n");	    
	    return 1;
	  }
          break;



        case 'l':
	  if (atof(optarg) > 0)
	    initial_tl = atof(optarg);
	  else {
	    fprintf (stderr, "Option -tl requires a positive argument.\n");	    
	    return 1;
	  }
          break;
     

        case 'h': 
	  printHelp(flag);
	  exit(0);	
	
	  // unrecognized options or options without argument
        case '?':
          return 1;
     

        default:
          abort ();

	} //end switch
    } //end while

  // double check the validity of arguments
  if (initial_rho == 0.0 && rho_flag == 1) {
    fprintf (stderr, "With rho_flag on, option -rho requires a positive argument.\n");	    
    return 1;
  }

  if (initial_gamma == 0.0 && gamma_flag == 1) {
    fprintf (stderr, "With gamma_flag on, option -gamma requires a positive argument.\n");	    
    return 1;
  }

  if (initial_tl == 0.0 && tl_flag == 1) {
    fprintf (stderr, "With tl_flag on, option -tl requires a positive argument.\n");	    
    return 1;
  }
  
  mle_flag = (rho_flag) || (gamma_flag) || (tl_flag);

  if(in_file == '\0') {
    fprintf (stderr, "Option -in is required\n");	    
    return 1;
  }

  command = "\nCommand line :";
  for(int i = 0; i < argc; ++i)
    command += (string(" ") + argv[i]);
  command += string("\n");

  // default random seed
  if(seed == 0) 
    seed = time(NULL);
 
  return 0;
  
} //end parsecommandline


void printHelp(int flag)
{
  stringstream usage, help, param, prereq, output;

  if(flag == LIKELIHOOD) {
    help << "\noverpaintLike" 
	 << " - compute the log likelihood\n"
	 << endl;

    usage << "Usage:\n"
          << "   overpaintLike ";
  } 
  else {
    help << "\noverpaintEst" 
	 << " - compute the MLE\n"
	 << endl;

    usage << "Usage:\n"
          << "   overpaintEst ";
  }

  usage << "-i[n] <in_file> "
        << "[-o[ut] <out_file>] "
        << "[-s[eed] <seed>] "
        << "[-[or]d[er]] <order_file>] "
        << "[-n[Order] <n_order>] "
        << "[-[re]p <n_rep>] ";

  param << "Parameters:\n"
        << "   -in   		Data file to use\n"
        << "   -out 		Output file (default: terminal)\n"
        << "   -seed		Seed for random permutations (default: time(NULL), ignored if -order specfied)\n"
        << "   -order		Ordering file to use (overrides -seed, -nOrder and -rep)\n"
        << "   -nOrder     		Number of random orderings (ignored if -order specfied)\n"
        << "   -rep			Number of repetitions (ignored if -order specfied)\n";


  if(flag == ESTIMATION) {
    usage << "[-m[le] <0/1 0/1 0/1>] ";
    param << "   -mle			MLE flag (three binary numbers) for rho, gamma and tl\n";
  }

  usage << "[-t[heta] <theta>] "
        << "[-r[ho] <rho>] "
        << "[-g[amma] <gamma>] "
        << "[-[t]l <tl>] "
        << "[-h]"
        << std::endl << std::endl;

  param << "   -theta		(Initial) Mutation rate (default: Watterson's estimator)\n"
        << "   -rho			(Initial) Crossover rate\n"
        << "   -gamma 		(Initial) Gene-conversion rate\n"
        << "   -tl 			(Initial) Mean gene-conversion tract length\n"	  
        << "   -h	       		This help\n"
        << std::endl;
 

  output << "Output:\n"
         << "   default:             output plain text to terminal,\n"
         << "   with -out:   	save in the file out_file,\n"
         << "   without -order:      save the random orderings in the in_file_orderxxx.txt.\n"
         << std::endl;

  cout << help.str() 
       << usage.str()
       << param.str()
       << output.str();	
}
