/**
 * PAC_GeneCo.cc
 *
 * Version :   0.2
 * Author  :   Junming Yin
 * 
 * (c) UC Berkeley, 2010
 * 
 */

#include <iostream>
#include <gsl/gsl_math.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_errno.h>
#include "PAC_GeneCo.h"

#define EPS     1.0e-30	                	// to scale the entries of doubleMatrix
#define INF     1.0E100    		    	// infinity 
#define MU      0.5
#define SIGMA   0.4                               

using namespace std;

// default constructor with explicit initializers
PAC_GeneCo::PAC_GeneCo(): rho_mle(0.0), gamma_mle(0.0), tl_mle(0.0), g00(0.0), g01(0.0), g1_11(0.0), g1_12(0.0), g10(0.0), 
                          g11(0.0), g12(0.0), g11_1(0.0), g11_11(0.0), g11_12(0.0), g12_1(0.0), g12_11(0.0), g12_12(0.0), 
                          g12_21(0.0), g12_13(0.0), allForwardsSum(0.0), currentRep(0), mode(0), funevals(0) {}


// constructor with arguments, avoid coping Data_info object with reference
PAC_GeneCo::PAC_GeneCo(const Data_info& d): data(d), rho_mle(0.0), gamma_mle(0.0), tl_mle(0.0), g00(0.0), g01(0.0), g1_11(0.0), g1_12(0.0), g10(0.0), 
                                            g11(0.0), g12(0.0), g11_1(0.0), g11_11(0.0), g11_12(0.0), g12_1(0.0), g12_11(0.0), g12_12(0.0), g12_21(0.0), 
                                            g12_13(0.0), allForwardsSum(0.0), currentRep(0), mode(0), funevals(0)
{ 
  int N = data.N;
  int L = data.L;

  // initialize emissionProb, independent of rho, gamma and tl
  emissionProb.resize(N - 1);
  for(int k = 1; k <= N - 1; k++) {
    emissionProb[k - 1].resize(2);
    // no mutation 
    emissionProb[k - 1][0] = (2 * (double) k + data.theta)/(2 * (double) k + 2 * data.theta);
    // mutation
    emissionProb[k - 1][1] = 1 - emissionProb[k - 1][0];
  }

  // initialize some temp variables
  vector<double> vd(N-1, 0.0);
  vector<double> vdn(N, 0.0);	
  doubleMatrix vvd(N-1, vdn);

  xForwardsSums = vd;
  gForwardsSums = vdn;
  singletForwardsLik.resize(L, vvd);

  gSums = vd;
  activeGSums = vd;
  gActiveGSums.resize(N-1, vd); 
  xGSums.resize(N-1,vd);
  xActiveGSums.resize(N-1,vd);
  vector<doubleMatrix> vvvd(N-1, gActiveGSums);
  doubletForwardsLik.resize(L, vvvd);

  forwardsLik.resize(L, vd);
}


int PAC_GeneCo::find_mle(int nvars, double* startpt, double* endpt, double epsilon, int itermax) 
{

  // 1-D search
  if (nvars == 1) {
    const gsl_min_fminimizer_type *T = gsl_min_fminimizer_brent; 
    gsl_min_fminimizer *s;
    double lower = 0.0, upper = 15.0;
    gsl_function min_func;

    // number of iterations in optimization
    int iter = 0;
    int status;

    // starting point
    double pt = startpt[0];

    // initialize the members
    min_func.function = mle_objective;
    min_func.params = (void*)this;

    s = gsl_min_fminimizer_alloc (T);
    gsl_min_fminimizer_set (s, &min_func, pt, lower, upper);
   
    do {
        iter++;
        status = gsl_min_fminimizer_iterate (s);

        pt     = gsl_min_fminimizer_x_minimum (s);
        lower  = gsl_min_fminimizer_x_lower (s);
        upper  = gsl_min_fminimizer_x_upper (s);

        status = gsl_min_test_interval (lower, upper, 0.001, 0.0);
    }
    while (status == GSL_CONTINUE && iter < itermax);
        
    endpt[0] = pt;
    gsl_min_fminimizer_free (s);

    return iter;
    } 
  // multidimensional search
  else { 
    const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex;
    gsl_multimin_fminimizer *s = NULL;
    gsl_vector *step_size, *pt;
    gsl_multimin_function min_func;

    // number of iterations in optimization
    int iter = 0;
    int status;
    double size;
        
    // starting point
    pt = gsl_vector_alloc (nvars);
    for (int i = 0; i < nvars; i++)
      gsl_vector_set(pt, i, startpt[i]);
         
    // initial step size
    step_size = gsl_vector_alloc(nvars);
    gsl_vector_set_all(step_size, 1.0);
         
    // initialize the members
    min_func.n = nvars;
    min_func.f = mle_objective;
    min_func.params = (void*)this;
         
    s = gsl_multimin_fminimizer_alloc(T, nvars);
    gsl_multimin_fminimizer_set(s, &min_func, pt, step_size);
         
    do
      {
	iter++;
	status = gsl_multimin_fminimizer_iterate(s);
               
	if (status) 
	  break;
         
	size = gsl_multimin_fminimizer_size (s);
	status = gsl_multimin_test_size (size, epsilon);
      }
    while (status == GSL_CONTINUE && iter < itermax);

    for (int i = 0; i < nvars; i++)
      endpt[i] = gsl_vector_get (s->x, i);

    gsl_vector_free(pt);
    gsl_vector_free(step_size);
    gsl_multimin_fminimizer_free(s);

    return iter;
  }

}

// for 1-D only
double mle_objective(double v, void *params) 
{
  PAC_GeneCo* pPAC_GeneCo = (PAC_GeneCo*)params;
  double pt[1];
  pt[0] = v;
  return pPAC_GeneCo->objective(pt);
}


// for multidimensional search
double mle_objective(const gsl_vector *v, void *params)
{
  PAC_GeneCo* pPAC_GeneCo = (PAC_GeneCo*)params;
  int nvars = v->size;

  double pt[nvars];
  for(int i = 0; i< nvars; i++)
    pt[i] = gsl_vector_get(v, i);
  return pPAC_GeneCo->objective(pt);
}

double PAC_GeneCo::objective(double *pt)
{
  // increments the number of function evaluations
  funevals++;
  // return the negative likelihood
  if(mode == 0)   	// rho_flag = 1, gamma_flag = 0, tl_flag = 0
    return (-1.0 * logLike0(pt[0]));
  else {		  	
    int i = 0;
    double r = (rho_flag == 0)? initial_rho : pt[i++];
    double g = (gamma_flag == 0)? initial_gamma : pt[i++];
    double l = (tl_flag == 0) ? 1.0/initial_tl : pt[i++];

    // return regularized likelihood function
    return ( -1.0 * logLike(r, g, l) + log(l) + pow(log10(l) - MU , 2)/(2*pow(SIGMA,2)) ); 
  }
}


double PAC_GeneCo::logLike(double rho, double gamma, double lambda)
{
  // compute the log likelihood specified at rho, gamma and lambda
  // lambda: rate of terminating a gene conversion track (not the tract length in the paper)
  int N = data.N;
  int nOrder = data.nOrder;
  double sumLik = 0.0;	
  double constant = 0.0;


  if(rho < 0.0 || gamma <= 0.0 || lambda <= 0.0) {
    //	cout << "logLike: rho, gamma and lambda cannot be less than zero" << endl;
    return -1.0 * INF;
  }
	
  for(int ord = 0; ord < nOrder; ++ord) {
    double logLik = 0.0;  // or set it to -log(2) * L: the unconidional likelihood of the 1st haplotype
    // permute the data.haplotypes using data.orderings[currentRep][ord]
    currentOrder = data.orderings[currentRep][ord];
    // logLik is the loglikelihood for currentOrder
    for(int k = 1; k <= N - 1; ++k) 
      // compute the (k+1)th haplotype given the first k haplotypes
      logLik += cond_logLike(k, rho, gamma, lambda);
	
    if (ord == 0)
      constant = logLik;
    // log(sum(exp(logLik)))
    sumLik += exp(logLik - constant);
  } // end ord

  // average of loglikelihood (across different orderings) for this particular repetition
  return (log(sumLik) + constant - log((double) nOrder));
}


double PAC_GeneCo::cond_logLike(int k, double rho, double gamma, double lambda) 
{
  // permute the data.haplotypes using currentOrder
  // then compute the (k + 1)th haplotype given the first k haplotypes, k = 1, ..., data.N - 1
  // use data.haplotypes[currentOrder[k]-1] and emissionProb[k-1]

  int L = data.L;

  /************************************
   initial probabilities of G/X chain
  ************************************/
  double distance = data.loci[1].dist; 
  vector<double> initial_G = gInitialProb(k, distance, gamma, lambda);
  double initial_X = 1.0/(double)k;


  /**********************************************************************************************************
                               initialize singletForwardsLik[0] (first locus) 
   indexed as: singletForwardsLik[locus][xstate-1][gstate], 0<=locus<=(L-1); 1<=xstate<=N-1; 0<=gstate<=N-1
  **********************************************************************************************************/

  int mut, copyFrom;
  for(int xstate = 1; xstate <= k; ++xstate) { 
    // gstate == 0 (null state), the active state is xstate
    mut = 1;
    if(data.haplotypes[currentOrder[k]-1][0] == data.haplotypes[currentOrder[xstate-1]-1][0])
      mut = 0;        
    singletForwardsLik[0][xstate-1][0] = initial_G[0] * initial_X * emissionProb[k-1][mut];

    // gstate != 0, the active state is gstate
    for(int gstate = 1; gstate <= k; ++gstate) {
      mut = 1;
      if(data.haplotypes[currentOrder[k]-1][0] == data.haplotypes[currentOrder[gstate-1]-1][0])
        mut = 0;
      singletForwardsLik[0][xstate-1][gstate] = (initial_G[1]/k) * initial_X * emissionProb[k-1][mut];
    }
  }
  /**************************************************************************************************************************************
                                        initialize doubletForwardsLik[0] (first locus) 
   indexed as: doubletForwardsLik[locus][xstate-1][gstate-1][activeGstate-1], 0<=locus<=(L-1); 1<=xstate<=N-1; 1<=gstate, activeGstate<=(N-1)
  ***************************************************************************************************************************************/
  for(int gstate = 1; gstate <= k; ++gstate) {
    for(int activeGstate = 1; activeGstate <= k; ++activeGstate) {
      // the active state is activeGstate
      mut = 1;
      if(data.haplotypes[currentOrder[k]-1][0] == data.haplotypes[currentOrder[activeGstate-1]-1][0])
        mut = 0;

      if(gstate == activeGstate)  { // doublet with the same state
	for(int xstate = 1; xstate <= k; ++xstate)
	  doubletForwardsLik[0][xstate-1][gstate-1][activeGstate-1] = (initial_G[2]/k) * initial_X * emissionProb[k-1][mut];
      }
      else { // doublet with the different state
	for(int xstate = 1; xstate <= k; ++xstate)
	  doubletForwardsLik[0][xstate-1][gstate-1][activeGstate-1] = (initial_G[3]/(k*k - k)) * initial_X * emissionProb[k-1][mut];
      }
    }
  }

  /**********************************************************************************************************
                fill in singletForwardsLik[loc] and doubletForwardsLik[loc]  (1<=locus<=(L-1)) 
  **********************************************************************************************************/
  double xTransition, xNoTransition;
  double sumWithoutRecomb, sumWithRecomb;

  double max = 0.0, doubleMax = 0.0;
  double logfactor = 0.0, lik=0.0;

  for(int loc = 1;loc <= L - 1; loc++)  {
    distance = data.loci[loc].dist;
    // first find transition probabilities of X chains for this locus.
    xTransition = xTransitionProb(k, distance, rho, 1);
    xNoTransition = xTransitionProb(k, distance, rho, 0);

    // then find transition probabilities of G chains for this locus.
    gTransitionProb(k, distance, gamma, lambda);

    // make partial sums of previous locus index by (loc-1)
    singletMakePartialSums(k, loc);
    doubletMakePartialSums(k, loc);

		
    for(int xstate = 1; xstate <= k; ++xstate)  {  		// xstate = 1,...,k, xindex = xstate - 1
      /******************************************************************
          fill in singletForwardsLik[loc][xstate-1] 
      ******************************************************************/
      for(int gstate = 0; gstate <= k ; ++gstate)  	{	// gstate = 0,1,...,k. gindex = gstate
        if(gstate == 0)  {	  	//  null state                			
          //coming from states with same xstate, and gstate' != 0
          sumWithoutRecomb = xForwardsSums[xstate-1] * g10 * xNoTransition; 

          //coming from state with same xstate, and gstate' = 0
          sumWithoutRecomb += singletForwardsLik[loc-1][xstate-1][0] * g00 * xNoTransition;

          //coming from states with different xstate, and gstate' != 0
          sumWithRecomb = (allForwardsSum - xForwardsSums[xstate-1] - gForwardsSums[0]) * g10 * xTransition; 

          //coming from states with different xstate, and gstate' = 0
          sumWithRecomb += (gForwardsSums[0] - singletForwardsLik[loc-1][xstate-1][0]) * g00 * xTransition;
	  
        } 	// end if
        else {  						//  gstate = 1,...,k, gindex = gstate 
          //coming from states with same xstate but gstate' = 0
          sumWithoutRecomb = singletForwardsLik[loc-1][xstate-1][0] * g01;

          //coming from states wtih same xstate and same gstate
          sumWithoutRecomb += singletForwardsLik[loc-1][xstate-1][gstate] * g11;

          //coming from states with same xstate and different (non-zero) gstate'
          sumWithoutRecomb += (xForwardsSums[xstate-1] - singletForwardsLik[loc-1][xstate-1][gstate]) * g12;

          // coming from [gstate,gstate]
          sumWithoutRecomb += doubletForwardsLik[loc-1][xstate-1][gstate-1][gstate-1] * g11_1;

          // coming from [gstate,gstate'] or [gstate',gstate]
          sumWithoutRecomb += (xGSums[xstate-1][gstate-1] + xActiveGSums[xstate-1][gstate-1] - 2 * doubletForwardsLik[loc-1][xstate-1][gstate-1][gstate-1]) * g12_1;

          sumWithRecomb = 0.0;
        } //end else

        singletForwardsLik[loc][xstate-1][gstate] = sumWithRecomb + sumWithoutRecomb;

        mut = 1;
        copyFrom = (gstate == 0)? xstate : gstate;
        if(data.haplotypes[currentOrder[k]-1][loc] == data.haplotypes[currentOrder[copyFrom-1]-1][loc])
          mut = 0;

        singletForwardsLik[loc][xstate-1][gstate] *= emissionProb[k-1][mut];

        if(singletForwardsLik[loc][xstate-1][gstate] > max)
          max = singletForwardsLik[loc][xstate-1][gstate];
      } //end gstate 

      /******************************************************************
	  	    finish filling in singletForwardsLik[loc][xstate-1] 
      ******************************************************************/

      /******************************************************************
	            fill in doubletForwardsLik[loc][xstate-1] 
      ******************************************************************/

      for(int gstate = 1; gstate <= k ; ++gstate)  	{				// gstate = 1,...,k
        for(int activeGstate=1; activeGstate<=k; ++activeGstate)  {		    	// activeG = 1,...,k	

          if(gstate == activeGstate)   {						// diagonal elements
            // from [gstate]
            sumWithoutRecomb = singletForwardsLik[loc-1][xstate-1][gstate] * g1_11;  	
            // from [gstate,gstate]
            sumWithoutRecomb += doubletForwardsLik[loc-1][xstate-1][gstate-1][activeGstate-1] * g11_11;
            // from [gstate,gstate'] or [gstate',gstate]
            sumWithoutRecomb += (xGSums[xstate-1][gstate-1] + xActiveGSums[xstate-1][gstate-1] - 2 * doubletForwardsLik[loc-1][xstate-1][gstate-1][gstate-1]) * g12_11; 

            sumWithRecomb = 0.0;
          }
          else {									// off-diagonal elements
            // from [gstate,activeGstate]
            sumWithoutRecomb = doubletForwardsLik[loc-1][xstate-1][gstate-1][activeGstate-1] * g12_12; 
            // from [activeGstate,gstate]
            sumWithoutRecomb += doubletForwardsLik[loc-1][xstate-1][activeGstate-1][gstate-1] * g12_21;	

            // from [gstate,gstate']
            sumWithoutRecomb += (xGSums[xstate-1][gstate-1] - doubletForwardsLik[loc-1][xstate-1][gstate-1][gstate-1] - doubletForwardsLik[loc-1][xstate-1][gstate-1][activeGstate-1]) * g12_13; 	
            // from [gstate',gstate]
            sumWithoutRecomb += (xActiveGSums[xstate-1][gstate-1] - doubletForwardsLik[loc-1][xstate-1][gstate-1][gstate-1] - doubletForwardsLik[loc-1][xstate-1][activeGstate-1][gstate-1]) * g12_13; 	
	    
            // from [activeGstate,gstate']
            sumWithoutRecomb += (xGSums[xstate-1][activeGstate-1] - doubletForwardsLik[loc-1][xstate-1][activeGstate-1][gstate-1] - doubletForwardsLik[loc-1][xstate-1][activeGstate-1][activeGstate-1]) * g12_13;
            // from [gstate',activeGstate]
            sumWithoutRecomb += (xActiveGSums[xstate-1][activeGstate-1] - doubletForwardsLik[loc-1][xstate-1][gstate-1][activeGstate-1] - doubletForwardsLik[loc-1][xstate-1][activeGstate-1][activeGstate-1])* g12_13;

            // from [gstate,gstate] or [activeGstate,activeGstate]
            sumWithoutRecomb += (doubletForwardsLik[loc-1][xstate-1][gstate-1][gstate-1] + doubletForwardsLik[loc-1][xstate-1][activeGstate-1][activeGstate-1]) * g11_12; 					
            // from [gstate] or [activeGstate], different index
            sumWithoutRecomb += (singletForwardsLik[loc-1][xstate-1][gstate] + singletForwardsLik[loc-1][xstate-1][activeGstate]) * g1_12; 	
   
            sumWithRecomb = 0.0;
          } // end else

          doubletForwardsLik[loc][xstate-1][gstate-1][activeGstate-1] = sumWithRecomb + sumWithoutRecomb;

          mut = 1;
          copyFrom = activeGstate;
          if(data.haplotypes[currentOrder[k]-1][loc] == data.haplotypes[currentOrder[copyFrom-1]-1][loc])
            mut=0; 	//our allele is the same as our nearest neighbour

          doubletForwardsLik[loc][xstate-1][gstate-1][activeGstate-1] *= emissionProb[k-1][mut];

          if(doubletForwardsLik[loc][xstate-1][gstate-1][activeGstate-1]  > doubleMax)
            doubleMax = doubletForwardsLik[loc][xstate-1][gstate-1][activeGstate-1];
        } 		//end activeGstate
      }  	//end gstate, finish fill out the doubleForwardsLik

      /******************************************************************
	  	    finish filling in doubletForwardsLik[loc][xstate-1] 
      ******************************************************************/
    } //end xstate

      /******************************************************************
	     rescale entries of doubletForwardsLik and singletForwardsLik
      ******************************************************************/

    max = (max > doubleMax)? max: doubleMax;

    if(max < EPS)  {
      logfactor += log(max);

      for(int xstate = 1; xstate <= k; ++xstate)    							// xstate = 1,...,k, xindex = xstate - 1
        for(int gstate = 0; gstate <= k ; ++gstate)  							// gstate = 0,1,...,k, gindex = gstate
          singletForwardsLik[loc][xstate-1][gstate] /= max;

      for(int xstate = 1; xstate <= k; ++xstate)    							// xstate = 1,...,k, xindex = xstate - 1
        for(int gstate = 1; gstate <= k ; ++gstate)  							// gstate = 1,...,k, gindex = gstate - 1 
          for(int activeGstate=1; activeGstate<=k; ++activeGstate)	   				// activeGstate = 1,...,k, activeGindex = activeGstate - 1
            doubletForwardsLik[loc][xstate-1][gstate-1][activeGstate-1] /= max;
                
    }
    doubleMax = 0.0;
    max = 0.0;

  } //end loci	

  /******************************************************************
   compute the likelihood by summing over the last slice of matrices
  ******************************************************************/
  for(int xstate = 1; xstate <= k; ++xstate)  {  					// xstate = 1,...,k, xindex = xstate - 1
    for(int gstate = 0; gstate <= k ; ++gstate)  					// gstate = 0,1,...,k. gindex = gstate
      lik += singletForwardsLik[L-1][xstate-1][gstate];

    for(int gstate = 1; gstate <= k ; ++gstate)  				    	// gstate = 1,...,k, gindex = gstate - 1
      for(int activeGstate=1; activeGstate<=k; ++activeGstate)	   			// activeGstate = 1,...,k, activeGindex = activeGstate - 1
        lik += doubletForwardsLik[L-1][xstate-1][gstate-1][activeGstate-1];
  }

  return log(lik) + logfactor;
}


double PAC_GeneCo::xTransitionProb(int k, double distance, double rho, int changeX)
{
  double xTranProb = 0.0;

  if(changeX == 0)  // no change
    xTranProb = 1.0/(double) k + ((double) k - 1)/(double) k * exp(-rho*distance/(double) k);
  else  { 
    if (k == 1)
      xTranProb = 0.0;
    else
      xTranProb = (1.0 - exp(-rho*distance/(double) k))/(double) k;
  }
  return xTranProb;
}

void PAC_GeneCo::gTransitionProb(int k, double distance, double gamma, double lambda)	
{

  long double expmdL = exp(-distance*lambda);
  long double expmdL2 = exp(-2.0*distance*lambda);
  long double dL = distance * lambda;
  long double kL = (long double) k * lambda;
  long double k2L = (long double) k * (long double) k * lambda;

  g00 = (kL + gamma * (-1.0 + expmdL + dL));
  g01 = (1.0 - expmdL)* (gamma/(double) k);

  g1_11 = gamma * (expmdL - expmdL2);
  g1_12 = (k < 2)? 0: (g1_11/2.0);
  g10 = k2L * (1.0 - expmdL);
  g11 = (k2L * expmdL + gamma * (-expmdL + 1 - dL * expmdL + (k + 1.0) * (-expmdL + expmdL2 + dL * expmdL)));
  g12 = (k < 2)? 0:(gamma * (-2.0 * expmdL + 1 + expmdL2));

  g11_1 = 2.0 * k2L * (1.0 - expmdL);
  g11_11 = (k2L * expmdL + 2.0 * gamma * (-expmdL + 1 - dL * expmdL));
  g11_12 = (k < 2)? 0:(gamma * (-expmdL + 1 - dL * expmdL));

  // only non-zeros when k>=2
  g12_1 = (k < 2)? 0:(k2L * (1.0 - expmdL));
  g12_11 = (k < 2)? 0:( gamma * (-expmdL + 1 - dL * expmdL));
  g12_21 = (k < 2)? 0:g12_11;
  g12_12 = (k < 2)? 0:(g12_11 + k2L * expmdL);
  g12_13 = (k < 3)? 0:(g12_11/2.0);

  // need to normalize the probability
  long double normalized = (k * g01 + g00);
  if(normalized != 0.0) {
    g00 /= normalized;
    g01 /= normalized;
  }

  normalized = g1_11 + 2.0*(k-1)*g1_12 + g11 + (k-1)*g12 + g10;
  if(normalized != 0.0) {
    g1_11 /= normalized;
    g1_12 /= normalized;
    g11 /= normalized;
    g12 /= normalized;
    g10 /= normalized;
  }

  normalized = g11_1 + g11_11 + 2.0*(k-1)*g11_12;
  if(normalized != 0.0) {
    g11_1 /= normalized;
    g11_11 /= normalized;
    g11_12 /= normalized;
  }

  normalized = 2.0*g12_11 + g12_12 + g12_21 + 4.0*(k-2)*g12_13 + 2.0*g12_1;
  if(normalized != 0.0) {
    g12_1 /= normalized;
    g12_12 /= normalized;
    g12_21 /= normalized;
    g12_13 /= normalized;
    g12_11 /= normalized;
  } 
}

vector<double> PAC_GeneCo::gInitialProb(int k, double distance, double gamma, double lambda) {

  /****************************
       local variables 
  ****************************/

  vector<double> initial_vector(4, 0.0);
  doubleMatrix collapsedTransitionP(4, initial_vector);

  /****************************
   ****************************/
  gTransitionProb(k, distance, gamma, lambda);

  // 0 = 0, 1 = g, 2 = gg, 3 = gg'
  collapsedTransitionP[0][0] = g00;
  collapsedTransitionP[0][1] = k*g01;
  collapsedTransitionP[1][0] = g10; 
  collapsedTransitionP[1][1] = g11+(k-1)*g12; 
  collapsedTransitionP[1][2] = g1_11; 
  collapsedTransitionP[1][3] = 2.0*(k-1)*g1_12; 
  collapsedTransitionP[2][1] = g11_1; 
  collapsedTransitionP[2][2] = g11_11; 
  collapsedTransitionP[2][3] = 2.0*(k-1)*g11_12; 
  collapsedTransitionP[3][1] = 2.0*g12_1; 
  collapsedTransitionP[3][2] = 2.0*g12_11; 
  collapsedTransitionP[3][3] = g12_12+g12_21+4.0*(k-2)*g12_13; 

  // format collapsedTransitionP into a vector for gsl
  double temp[16] = {0};
  // have been transposed here since we need the LEFT eigenvector
  for(int i = 0;i<16; i++)
    temp[i] = collapsedTransitionP[(i%4)][(i/4)]; 


  gsl_matrix_view m = gsl_matrix_view_array (temp, 4, 4);
  
  gsl_vector_complex *eval = gsl_vector_complex_alloc (4);
  gsl_matrix_complex *evec = gsl_matrix_complex_alloc (4, 4);
     
  gsl_eigen_nonsymmv_workspace *w = gsl_eigen_nonsymmv_alloc (4);
       
  gsl_eigen_nonsymmv (&m.matrix, eval, evec, w);
     
  gsl_eigen_nonsymmv_free (w);
     
  gsl_eigen_nonsymmv_sort (eval, evec, GSL_EIGEN_SORT_ABS_DESC);
  
  // get the (smallest) LEFT eigenvector that corresponds to eigenvalue 1
  gsl_vector_complex_view evec_0 = gsl_matrix_complex_column (evec, 0);
  double norm_v = 0.0;
  for (int j = 0; j < 4; j++) {
    gsl_complex z = gsl_vector_complex_get(&evec_0.vector, j);
    initial_vector[j] = fabs(GSL_REAL(z));
    norm_v += initial_vector[j];
  }

  // normalize the eigenvector
  for (int j = 0; j < 4; j++)
    initial_vector[j] /= norm_v; 

  gsl_vector_complex_free(eval);
  gsl_matrix_complex_free(evec);

#ifdef DEBUG
  // using power method to check the consistency
  vector<double> vector(4,0.0);
  vector[0] = 1.0;
  double newVector[4] = {1.0,0.0,0.0,0.0};
  double maxChange = 1.0;
  int counter = 0;
  while(maxChange > 1e-10){
    maxChange = 0.0;
    counter ++;
    for(int i = 0; i< 4; i++){
      newVector[i] = 0.0;
      for(int j = 0; j < 4; j++){
	newVector[i] += vector[j]*collapsedTransitionP[j][i];
      }
      if(fabs(newVector[i]-vector[i]) > maxChange) maxChange = fabs(newVector[i]-vector[i]);
    }
    for(int i = 0; i< 4; i++) 
      vector[i] = newVector[i];
  }
  return vector;
#endif

  return initial_vector;
}


void PAC_GeneCo::singletMakePartialSums(int k, int loc) 
{
  // make partial sums for singletForwardsLik[loc-1]
  int xstate, gstate;
  double entry;
  
  // set all partial sums to be 0.0
  allForwardsSum = 0.0;

  for(xstate = 1; xstate <= k; ++xstate)  		// xstate = 1,...,k, xindex = xstate - 1
    xForwardsSums[xstate-1] = 0.0;

  for(gstate = 0; gstate <= k; ++gstate) 		// gstate = 0,1,...,k. gindex = gstate
    gForwardsSums[gstate] = 0.0;

  // start iteration
  for(xstate = 1; xstate <= k; ++xstate)  		// xstate = 1,...,k, xindex = xstate - 1
    for(gstate = 0; gstate <= k; ++gstate) 	{   	// gstate = 0,1,...,k. gindex = gstate
      entry = singletForwardsLik[loc-1][xstate-1][gstate];
      if(gstate != 0)
        xForwardsSums[xstate - 1] += entry;
      gForwardsSums[gstate] += entry;
      allForwardsSum += entry;
    }
}

void PAC_GeneCo::doubletMakePartialSums(int k, int loc)
{
  // make partial sums for doubletForwardsLik[loc-1]
  int xstate, gstate, activeGstate;		
  double entry;

  // set all partial sums to be 0.0
  for(gstate=1; gstate<=k; ++gstate)   {						// gstate = 1,...,k
    gSums[gstate-1] = 0.0;
    activeGSums[gstate-1] = 0.0;
  }

  for(gstate=1; gstate<=k; ++gstate)							// gstate = 1,...,k
    for(activeGstate=1; activeGstate<=k; ++activeGstate)				// activeGstate = 1,...,k
      gActiveGSums[gstate-1][activeGstate-1] = 0.0;

  for(xstate = 1; xstate <= k; ++xstate)  						// xstate = 1,...,k
    for(gstate=1; gstate<=k; ++gstate)	 {						// gstate = 1,...,k
      xGSums[xstate-1][gstate-1] = 0.0;
      xActiveGSums[xstate-1][gstate-1] = 0.0;	
    }

  // start iteration
  for(xstate = 1; xstate <= k; ++xstate)  				 		// xstate = 1,...,k
    for(gstate=1; gstate<=k; ++gstate)							// gstate = 1,...,k
      for(activeGstate=1; activeGstate<=k; ++activeGstate)  {				// activeGstate = 1,...,k
        entry = doubletForwardsLik[loc-1][xstate-1][gstate-1][activeGstate-1];
        gSums[gstate-1] += entry;
        activeGSums[activeGstate-1] += entry;

        gActiveGSums[gstate-1][activeGstate-1] += entry;

        xGSums[xstate-1][gstate-1] += entry;
        xActiveGSums[xstate-1][activeGstate-1] += entry;
      }
}


double PAC_GeneCo::logLike0(double rho)
{
  // compute the log likelihood specified at rho
  int N = data.N;
  int nOrder = data.nOrder;
  double sumLik = 0.0;	
  double constant = 0.0;

  if(rho <= 0.0) {
    //	cout << "logLike0: rho cannot be less than zero" << endl;
    return -1.0 * INF;
  }

  for(int ord = 0; ord < nOrder; ord++) {
    double logLik = 0.0;  // or set it to -log(2) * L: the unconidional likelihood of the 1st haplotype
    // permute the data.haplotypes using data.orderings[currentRep][ord]
    currentOrder = data.orderings[currentRep][ord];
    for(int k = 1; k <= N - 1; k++) 
      // compute the (k+1)th haplotype given the first k haplotypes
      logLik += cond_logLike0(k, rho);
    if (ord == 0)
      constant = logLik;
    // log(sum(exp(logLik)))
    sumLik += exp(logLik - constant);
  } // end ord
  return (log(sumLik) + constant - log((double) nOrder));
}


double PAC_GeneCo::cond_logLike0(int k, double rho)  
{
  // permute the data.haplotypes using currentOrder
  // then compute the (k + 1)th haplotype given the first k haplotypes, k = 1, ..., data.N - 1
  // use data.haplotypes[currentOrder[k]-1] and emissionProb[k-1]

  /**********************************************************************************************************
                                   initialize forwardsLik[0] (first locus) 
             indexed as: forwardsLik[locus][xstate-1], 0<=locus<=(L-1); 1<=xstate<=N-1;
  **********************************************************************************************************/
  // initial probabilities of X chain
  double initial_X = 1.0/(double)k;

  int mut;
  for(int xstate = 1; xstate <= k; ++xstate) { 
    mut = 1;
    if(data.haplotypes[currentOrder[k]-1][0] == data.haplotypes[currentOrder[xstate-1]-1][0])
      mut = 0;
    forwardsLik[0][xstate-1] = initial_X * emissionProb[k-1][mut];
  }

  /***************************************************************
                fill in forwardsLik[loc] (1<=locus<=(L-1)) 
  ***************************************************************/
  double xTransition, xNoTransition;
  double distance;
  double max = 0.0, logfactor = 0.0, lik = 0.0;
  int L = data.L;

  for(int loc = 1;loc <= L - 1; loc++)  {
    distance = data.loci[loc].dist;
    xTransition = xTransitionProb(k, distance, rho, 1);
    xNoTransition = xTransitionProb(k, distance, rho, 0);

    // make partial sums of previous locus index by (loc-1)
    double forwardsSums = 0.0;
    for(int xstate = 1; xstate <= k; ++xstate)  	    	// xstate = 1,...,k
      forwardsSums += forwardsLik[loc-1][xstate-1];
	
    for(int xstate = 1; xstate <= k; ++xstate) {	 	// xstate = 1,...,k
      forwardsLik[loc][xstate-1] = (forwardsSums - forwardsLik[loc-1][xstate-1]) * xTransition + forwardsLik[loc-1][xstate-1] * xNoTransition;
      mut = 1;
      if(data.haplotypes[currentOrder[k]-1][loc] == data.haplotypes[currentOrder[xstate-1]-1][loc])
        mut = 0;
      forwardsLik[loc][xstate-1] *= emissionProb[k-1][mut];
		
      if(forwardsLik[loc][xstate-1] > max)
        max = forwardsLik[loc][xstate-1];
    } 	// end xstate

    /*********************************
         rescale entries of forwardsLik 
    *********************************/
    if(max < EPS)  {
      logfactor += log(max);
      for(int xstate = 1; xstate <= k; ++xstate)    		// xstate = 1,...,k, xindex = xstate - 1
        forwardsLik[loc][xstate-1] /= max;
    }
    max = 0.0;
  }   // end loc

  /******************************************************************
   compute the likelihood by summing over the last slice of matrices
  ******************************************************************/
  for(int xstate = 1; xstate <= k; ++xstate)   			// xstate = 1,...,k, xindex = xstate - 1
    lik += forwardsLik[L-1][xstate-1];

  return log(lik) + logfactor;
}


