/**
 * PAC_GeneCo.h
 *
 * Version :   0.2
 * Author  :   Junming Yin
 * 
 * (c) UC Berkeley, 2010
 * 
 */

#ifndef GUARD_PAC_GeneCo
#define GUARD_PAC_GeneCo

#include "Data_info.h"
#include <gsl/gsl_multimin.h>

typedef std::vector<std::vector<double> >  doubleMatrix;

class PAC_GeneCo {
 private:
  // the main data object 
  Data_info data; 
  // MLE of three fundamental parameters rho, gamma and tl
  std::vector<double> rho_mle;
  std::vector<double> gamma_mle;
  std::vector<double> tl_mle;
  // emission probabilities, independent of rho, gamma and tl, size of (data.N - 1)
  doubleMatrix emissionProb;                  
  // transition probabilities of G chain, depends on gamma and tl
  double g00, g01, g1_11, g1_12, g10, g11, g12, g11_1, g11_11, g11_12, g12_1, g12_11, g12_12, g12_21, g12_13;

  // some local variables to store partial sums in the HMM computation
  double allForwardsSum;
  std::vector<double> xForwardsSums;  
  std::vector<double> gForwardsSums;
  std::vector<doubleMatrix> singletForwardsLik;

  std::vector<double> gSums;
  std::vector<double> activeGSums;
  doubleMatrix gActiveGSums; 
  doubleMatrix xGSums;
  doubleMatrix xActiveGSums;
  std::vector<std::vector<doubleMatrix > > doubletForwardsLik;
  
  // for PAC model with only crossovers
  doubleMatrix forwardsLik;    		    
  
  // current index of repetitions and ordering
  int currentRep;                           
  std::vector<int> currentOrder;    
  // mode = 0: PAC with only crossovers; mode = 1: PAC with gene conversions
  int mode;			

 public:  
  // constructors
  PAC_GeneCo();				    
  PAC_GeneCo(const Data_info&); 	  
  
  // transition probabilities of X chain
  double xTransitionProb(int, double, double, int);

  // transition probabilities of G chain
  std::vector<double> gInitialProb(int, double, double, double);
  void gTransitionProb(int, double, double, double);

  // likelihood of PAC model with gene conversions specified at rho, gamma and lambda
  double logLike(double, double, double);  
  // conditional likelihoods of PAC model with gene conversions
  double cond_logLike(int, double, double, double); 
  // make partial sums for singletForwardsLik and doubletForwardsLik
  void singletMakePartialSums(int, int);
  void doubletMakePartialSums(int, int);

  // likelihood of PAC model with only crossovers specified at rho
  double logLike0(double);  
  // conditional likelihoods of PAC model with crossovers only
  double cond_logLike0(int, double);

  // number of function evaluations during optimization
  int funevals;  						
  // negative loglikelihood
  double objective(double *);   				
  // find MLE
  int find_mle(int, double*, double*, double, int);


  // accessory functions
  Data_info getData() const { return data; }
  void setCurrentRep(int r) { currentRep = r; }
  void setMode(int m) { mode = m; }
};

// global mle optimization
double mle_objective(const gsl_vector*, void*);

double mle_objective(double, void *);
#endif
