/**
 * overpaintEst.cc
 *
 * Version :   0.2
 * Author  :   Junming Yin
 * 
 * (c) UC Berkeley, 2010
 * 
 */


#include <iostream>		
#include <fstream>					
#include "Data_info.h"
#include "PAC_GeneCo.h"
#include "math.h"

#define  VARS		3	// max number of variables
#define  THRES		0.05	// threshold for optimization
#define  IMAX		1000	// max number of iterations 

using namespace std;


int main(int argc, char** argv)
{

  if (parseCommandline(argc, argv, ESTIMATION) == 1) 
    exit(1);

  // overpaintEst DOES need -mle option, and exclude -mle 0 0 0 and -mle 0 0 1
  if(rho_flag == 0 && gamma_flag == 0)  {
    cout << "overpaintEst (Error): option -mle is need." << endl;
    exit(1);
  }

  // global random number generator via gsl
  gsl_rng_env_setup();
  T = gsl_rng_default;
  rg = gsl_rng_alloc(T);
  gsl_rng_set(rg, seed);

  // create a Data_info object
  Data_info data(in_file, order_file, n_order, n_rep, initial_theta);
  initial_theta = data.theta;
  n_rep = data.nRep;
  n_order = data.nOrder;

  // output the orderings into infile_orderxxx.txt
  if(order_file == '\0') {
    char file[MAX_LENGTH];
    for(int i = 0; i < n_rep; ++i)  {
      sprintf(file, "%s_order%d.txt", in_file, i + 1);
      ofstream order_s(file);
      for(int j = 0; j < n_order; ++j) {
        for(int k = 0; k < data.N; ++k) 
          order_s << data.orderings[i][j][k]  << " ";
        order_s << endl;
      }
      order_s.close();
    }	
  }

  // create a PAC_GeneCo object, PAC_GeneCo prog(data);	
  PAC_GeneCo *prog = new PAC_GeneCo(data);


  string output = command; 
  output += string(command.size(), '-');
  output += string("\n"); 
  char line[MAX_LENGTH];


  // compute the MLE
  vector<double> vd(VARS, 0.0);
  doubleMatrix MLE(n_rep + 1, vd);  // including the initial values 
  double startpt[VARS];
  double endpt[VARS];
  vector<double> logLiks(n_rep, 0.0);
  vector<int> funevals(n_rep, 0);
  double sum_rho=0.0, sum_gamma=0.0, sum_ratio= 0.0, sum_lambda=0.0, sum_lik=0.0;
  int sum_funevals=0;


  // check validity of initial values of rho, gamma and tl
  if( initial_gamma == 0.0 ) {      
      // gamma_flag = 0, rho_flag = 1, initial_rho > 0, PAC with crossovers ONLY
      if ( initial_tl != 0) {
        cout << "overpaintEst (Warning): initial value of track length -tl is ignored" << endl;
        initial_tl = 0;
        tl_flag = 0;
      }
 
      sprintf(line, "Finding MLE for rho, starting at (%.5f, 0.0, 0.0) (averaged over %d orders)\n\n", initial_rho, n_order);
      output += string(line);
      sprintf(line, "%15s%13s%20s%14s\n", "repetition", "rho", "log-likelihood", "funevals");
      output += string(line);
      cout << output;

      // rho_flag = 1, gamma_flag = 0, tl_flag = 0
      prog->setMode(0);
      MLE[0][0] = initial_rho;
      for(int r = 0; r < n_rep; r++) {
        int nvars = 0;
        startpt[nvars++] = MLE[r][0];  	// the MLE from last repetition
 
        prog->funevals = 0; 		// reset the number of function evaluations
        prog->setCurrentRep(r);
        prog->find_mle(nvars, startpt, endpt, THRES, IMAX);  
        funevals[r] = prog->funevals;	// get the number of function evaluations
 
        MLE[r+1][0] = endpt[0];
        logLiks[r] = prog->logLike0(MLE[r+1][0]);
		
        sum_rho += MLE[r+1][0];
        sum_lik += logLiks[r];
        sum_funevals += funevals[r];
		
        // output the result to the terminal
        sprintf(line, "%15d %12.5f %19.5f %13d\n", r + 1, MLE[r+1][0], logLiks[r], funevals[r]);
        output += string(line);
        cout << line;
      } // end r

      sprintf(line, "\nAverage over %d repetitions:\n", n_rep);
      output += string(line);
      cout << line;
      sprintf(line, "%28.5f %19.5f %13d\n", sum_rho/n_rep, sum_lik/n_rep, sum_funevals/n_rep);
      output += string(line);
      cout << line;

  }  // end if
  else {    	                
      // initial_gamma > 0, gamma_flag = 0 or 1
    if (initial_tl == 0.0) {    // invalid
      cout << "overpaintEst: initial value of track length -tl has to be non-zero when -gamma is specified!" << endl;
      exit(1);
    }
    else {   
      // initial_tl > 0, tl_flag = 0 or 1, PAC with gene-conversions
	
      output += "Finding MLE for ";
      if(rho_flag != 0)
      	output += "rho, ";
      if(gamma_flag != 0)
       	output += "gamma, ";
      if(tl_flag != 0)
    	output += "tl, ";
      sprintf(line, "starting at (%.5f, %.5f, %.5f) (averaged over %d orders)\n\n", initial_rho, initial_gamma, initial_tl, n_order);
      output += string(line);
      sprintf(line, "%15s%13s%13s%14s%18s%20s%14s\n", "repetition", "rho", "gamma", "ratio", "tract-length", "log-likelihood", "funevals");
      output += string(line);
      cout << output;

      // initial_gamma != 0, initial_tl != 0
      prog->setMode(1);
      MLE[0][0] = initial_rho;
      MLE[0][1] = initial_gamma;
      MLE[0][2] = 1.0/initial_tl;   		// lambda (rate) instead of tract-length

      for(int r = 0; r < n_rep; r++) {
        int nvars = 0;
    	// the MLE from last repetition
        if(rho_flag == 1)
          startpt[nvars++] = MLE[r][0];  	
        if(gamma_flag == 1)
          startpt[nvars++] = MLE[r][1];
        if(tl_flag == 1)
          startpt[nvars++] = MLE[r][2];
	
        prog->funevals = 0; 			// reset the number of function evaluations
        prog->setCurrentRep(r);
        prog->find_mle(nvars, startpt, endpt, THRES, IMAX);  	
        funevals[r] = prog->funevals;		// get the number of function evaluations

        int i = 0;
        if(rho_flag == 1)
          MLE[r+1][0] = endpt[i++];
        else
          MLE[r+1][0] = MLE[r][0];
        if(gamma_flag == 1)
          MLE[r+1][1] = endpt[i++];
        else
          MLE[r+1][1] = MLE[r][1];
        if(tl_flag == 1)
          MLE[r+1][2] = endpt[i++];
        else
          MLE[r+1][2] = MLE[r][2]; 

        logLiks[r] = prog->logLike(MLE[r+1][0], MLE[r+1][1], MLE[r+1][2]);

        sum_rho += MLE[r+1][0];
        sum_gamma += MLE[r+1][1];
        sum_ratio += MLE[r+1][1]/MLE[r+1][0];
        sum_lambda += MLE[r+1][2];
        sum_lik += logLiks[r];
        sum_funevals += funevals[r];

        // output the result to the terminal
        sprintf(line, "%15d %12.5f %12.5f %13.5f %17.5f %19.5f %13d\n", r + 1, MLE[r+1][0], MLE[r+1][1], MLE[r+1][1]/MLE[r+1][0], 1.0/MLE[r+1][2], logLiks[r], funevals[r]);
        output += string(line);
        cout << line;
      } // end r		

      sprintf(line, "\nAverage over %d repetitions:\n", n_rep);
      output += string(line);
      cout << line;

      sprintf(line, "%28.5f %12.5f %13.5f %17.5f %19.5f %13d\n", sum_rho/n_rep, sum_gamma/n_rep, sum_gamma/sum_rho, n_rep/sum_lambda, sum_lik/n_rep, sum_funevals/n_rep);
      output += string(line);
      cout << line;
    }
  } // end else

  delete prog;

  // output the result into out_file
  if(out_file != '\0') {
    ofstream out_s(out_file);
    out_s << output;
    out_s.close();
  }

  return 0;
}

